/*
 * This code is part of the strategy game operational-space.
 * operational-space comes with ABSOLUTELY NO WARRANTY and is licensed under GPL-2.0.
 *    Copyright (C) 2024  VegOwOtenks, Sleppo04
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "FreeList.h"

#include <stdlib.h>

#include "../errorcodes.h"

// fl->first_free == fl->list.capacity
// => Nothing free

// FreeNode Pointer = 0 => No next pointer

static inline size_t _GetSizeT(void* source)
{
    size_t target;
    memcpy(&target, source, sizeof(target));
    return target;
}

static inline void _SetSizeT(void* destination, size_t value)
{
    memcpy(destination, &value, sizeof(value));
    return;
}

static void _FreeList_InitializePointers(FreeList* fl, size_t offset)
{
    for (size_t i = offset; i < DynamicArray_GetLength(&fl->list); i++) {
        // !! Possibly unaligned pointer
        size_t* element = DynamicArray_GetPointer(&fl->list, i);
        _SetSizeT(element, i + 1);
    }

    // !! Possibly unaligned pointer
    size_t* last = DynamicArray_GetPointer(&fl->list, DynamicArray_GetLength(&fl->list) - 1);
    _SetSizeT(last, fl->list.capacity);
}

int FreeList_Create(FreeList* target, size_t object_size, size_t initial_capacity, allocator_t* allocator)
{
    if (target == NULL) return EDESTADDRREQ;
    if (object_size == 0) return EINVAL;
    if (initial_capacity == 0) return EINVAL;

    // Guarantee we can always fit a size_t inside the element
    size_t element_size = object_size < sizeof(size_t) ? sizeof(size_t) : object_size;

    if (DynamicArray_Create(&target->list, element_size, initial_capacity, allocator)) {
        return ENOMEM;
    }
    target->list.reserved = target->list.capacity;
    target->object_size = object_size;

    // 0, init all
    _FreeList_InitializePointers(target, 0);
    target->first_free = 0;

    return EXIT_SUCCESS;
}

void FreeList_Destroy(FreeList* fl)
{
    if (fl == NULL) return;
    DynamicArray_Destroy(&fl->list);

    memset(fl, 0, sizeof(*fl));

    return;
}

int FreeList_Allocate(FreeList* fl, size_t* index)
{
    if (fl == NULL) {
        return EINVAL;
    }
    if (index == NULL) {
        return EDESTADDRREQ;
    }

    if (fl->first_free == fl->list.capacity) {
        if (DynamicArray_Resize(&fl->list, fl->list.capacity * 2)) {
            return ENOMEM;
        }
        fl->list.reserved = fl->list.capacity;
        _FreeList_InitializePointers(fl, fl->first_free);
    }

    *index = fl->first_free;
    fl->first_free = _GetSizeT(DynamicArray_GetPointer(&fl->list, fl->first_free));

    return EXIT_SUCCESS;
}

int FreeList_Free(FreeList* fl, size_t index)
{
    if (fl == NULL) {
        return EDESTADDRREQ;
    }
    if (index >= fl->list.capacity) {
        return EBOUNDS;
    }

#ifdef COMMON_FREELIST_CHECKFREE
    // Check whether the index is freed already

    size_t current = fl->first_free;

    // There are no free elements?
    if (fl->first_free == fl->list.capacity) {
        goto good_free;
    }

    // Iterate through the free-list
    do {
        if (current == index) {
            // Index is already in the free-list
            return EINVAL;
        }

        current = _GetSizeT(DynamicArray_GetPointer(&fl->list, current));
    } while (current != 0);

good_free:

#endif // CHECKFREE

    // Only `fl` itself may hold the 0 reference
    if (fl->first_free == 0) {
        // [head] [head_next] [...]

        size_t head = fl->first_free;
        // Where the first element points to
        size_t head_next = _GetSizeT(DynamicArray_GetPointer(&fl->list, fl->first_free));

        // Set the new element to point to the head_next
        _SetSizeT(DynamicArray_GetPointer(&fl->list, index), head_next);

        // Head points to the new element
        _SetSizeT(DynamicArray_GetPointer(&fl->list, head), index);

        // [head] [new] [head_next] [...]
    } else {
        // Point the new element to the previous first
        _SetSizeT(DynamicArray_GetPointer(&fl->list, index), fl->first_free);

        // Point the head to the new element
        fl->first_free = index;
    }

    return EXIT_SUCCESS;
}

int FreeList_FreeSorted(FreeList* fl, size_t index)
{
    if (fl == NULL) {
        return EDESTADDRREQ;
    }
    if (index >= fl->list.capacity) {
        return EBOUNDS;
    }

    size_t current  = fl->first_free;
    size_t previous = fl->first_free;

    while (current != fl->list.capacity && current < index) {
        previous = current;
        current = _GetSizeT(DynamicArray_GetPointer(&fl->list, current));
    }

#ifdef COMMON_FREELIST_CHECKFREE
    if (current == index) {
        return EINVAL;
    }
#endif

    if (current == previous) {
        // No iteration done, insert before first
        _SetSizeT(DynamicArray_GetPointer(&fl->list, index), current);
        fl->first_free = index;
    } else {
        _SetSizeT(DynamicArray_GetPointer(&fl->list, index), current);
        _SetSizeT(DynamicArray_GetPointer(&fl->list, previous), index);
    }

    return EXIT_SUCCESS;
}

int FreeList_Load(FreeList* fl, void* destination, size_t index)
{
    if (fl == NULL) return EINVAL;
    if (destination == NULL) return EDESTADDRREQ;
    if (index >= fl->list.capacity) return EBOUNDS;

    memcpy(destination, DynamicArray_GetPointer(&fl->list, index), fl->object_size);

    return EXIT_SUCCESS;
}

int FreeList_Store(FreeList* fl, size_t index, void* source)
{
    if (fl == NULL) return EINVAL;
    if (source == NULL) return EDESTADDRREQ;
    if (index >= fl->list.capacity) return EBOUNDS;

    memcpy(DynamicArray_GetPointer(&fl->list, index), source, fl->object_size);

    return EXIT_SUCCESS;
}

void* FreeList_GetPointer(FreeList* fl, size_t index)
{
    if (fl == NULL) return NULL;
    if (index >= fl->list.capacity) return NULL;

    return DynamicArray_GetPointer(&fl->list, index);
}
