/*
 * This code is part of the strategy game operational-space.
 * operational-space comes with ABSOLUTELY NO WARRANTY and is licensed under GPL-2.0.
 *    Copyright (C) 2024  VegOwOtenks, Sleppo04
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#ifndef DYNAMICBUFFER_H
#define DYNAMICBUFFER_H

#include <stdlib.h>
#include <string.h>

#include "../errorcodes.h"

#include "../allocator-interface/allocator-interface.h"

typedef struct DynamicBuffer {
    void*  array;
    size_t capacity;
    size_t used;

    allocator_t* allocator;
} dynamic_buffer_t;

/// @brief Create a new Dynamic Buffer at destination with initialCapacity initialCapacity
/// @param destination where the buffer will be stored
/// @param initialCapacity what it's initialCapacity will be
/// @return EINVAL, EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicBuffer_Create(dynamic_buffer_t* destination, size_t initialCapacity);


/// @brief Create a new Dynamic Buffer at destination with initialCapacity initialCapacity
/// @param destination where the buffer will be stored
/// @param initialCapacity what it's initialCapacity will be
/// @return EINVAL, EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicBuffer_CreateWithAllocator(dynamic_buffer_t* destination, size_t initialCapacity, allocator_t* allocator);

/// @brief This function checks that there are at least needed_unused free bytes in the allocated area
/// @param buffer buffer to check
/// @param needed_unused required free array size
/// @return EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicBuffer_EnsureUnusedCapacity(dynamic_buffer_t *buffer, size_t needed_unused);

/// @brief This function resizes the buffer to minimal_capacity, if necessary
/// @param buffer buffer to check
/// @param minimal_capacity minimal capacity of the buffer array
/// @return EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicBuffer_EnsureCapacity(dynamic_buffer_t *buffer, size_t minimal_capacity);

/// @brief This function resizes the buffers array via realloc to new_capacity
/// @param buffer buffer to resize
/// @param new_capacity capacity of the new buffer array
/// @return EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicBuffer_Resize(dynamic_buffer_t *buffer, size_t new_capacity);

/// @brief This function sets the buffers capacity to what it uses
/// @param buffer buffer to prune
/// @return EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicBuffer_Prune(dynamic_buffer_t* buffer);

/// @brief Resets the count of used bytes to "clear" the buffer
/// @param buffer buffer to reset
/// @return EINVAL, EXIT_SUCCESS
int DynamicBuffer_Reset(dynamic_buffer_t *buffer);

/// @brief Rewind the buffer pointer by bytes bytes, losing the bytes stored
/// @param buffer buffer to rewind
/// @param bytes How many bytes the buffer will lose
/// @return EINVAL, EBOUNDS, EXIT_SUCCESS
int DynamicBuffer_RewindBytes(dynamic_buffer_t* buffer, size_t bytes);

/// @brief Stores data[.data_size] at the end of the buffer array and resizes if necessary
/// @param buffer buffer in which the data will be stored
/// @param data data to store
/// @param data_size how many bytes will be copied from data
/// @return EINVAL, EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicBuffer_Store(dynamic_buffer_t *buffer, const void *data, size_t data_size);

/// @brief Calculate how many blocks are currently in the buffer
/// @param buffer buffer to query
/// @param block_size what's the size of a single block
/// @return How many of these block_sizes does the buffer currently hold
size_t DynamicBuffer_GetBlockCount(dynamic_buffer_t* buffer, size_t block_size);

/// @brief Get a pointer reference to the buffer contents at offset
/// @param buffer buffer to query
/// @param offset offset from the buffer start
/// @return Pointer to the address in the buffer at offset or NULL if out of Bounds
void* DynamicBuffer_ReadAt(dynamic_buffer_t* buffer, size_t offset);

/// @brief Get a pointer reference to indexn'th block in buffer
/// @param buffer buffer to query
/// @param block_size size of a single block
/// @param index which block to read
/// @return Pointer to the block at index or NULL if out of bounds
void* DynamicBuffer_ReadBlockAt(dynamic_buffer_t* buffer, size_t block_size, size_t index);

/// @brief Destroys the dynamic buffer and releases all resources held, the struct will not hold anything it did before
/// @param buffer buffer that shall be destroyed
/// @return EINVAL, EXIT_SUCCESS
int DynamicBuffer_Destroy(dynamic_buffer_t *buffer);

#endif
